//
// ini.cpp:
//	Implementation of classes InitFile, Section,
//	Key, AbsPtr, and PointerArray.
//
// Purpose:
//	Reads Windows-style .INI files into memory and provides
//	methods to access and update the values.  For more
//	information, see INI.DOC.
//
// Author(s): Mark Pflaging   C-serve  70233,1552
// Copyright 1992   Mark Thomas Pflaging
//
// Date:    5/15/92 - 9/21/92
//
#include "ini.hpp"
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE 1
#endif

#define INI_LINESIZE 254

void streamIgnore(istream & ini)
{
	unsigned char buffer[INI_LINESIZE + 1];

	while (!ini.eof()) {
		ini.getline(buffer, INI_LINESIZE);
		if (ini.gcount() < (INI_LINESIZE - 1))
			return;
	}
}

char * getline(istream &ini, char * buffer)
{
	ini.getline(buffer, INI_LINESIZE);
	// Strip trailing spaces.
	char *temp = buffer + strlen(buffer) - 1;
	while (isspace(*temp)) -- temp;
	*(temp + 1) = '\0';
	return temp;
}

PointerArray * getComments(istream & ini)
{
	PointerArray * retval = NULL;
	unsigned char buffer[INI_LINESIZE + 1];
	char * value;

	while (!ini.eof()) {
		ini >> buffer[0];
		if (buffer[0] == ';') {
			getline(ini, &buffer[1]);
			if (!retval) retval = new PointerArray(2);
			retval->insertString(buffer);
		}
		else {
			ini.putback(buffer[0]);
			return retval;
		}
	}
	return retval;
}

void dumpComment(AbsPtr & element, void * where)
{
	ostream & dest = *((ostream *)where);
	dest << (char *)(element.get()) << endl;
}

ostream & dumpComments(ostream & where, PointerArray * comments)
{
	if (comments) comments->doAll(dumpComment, &where);
	return where;
}

#pragma argsused
void killComment(AbsPtr & element, void * dummy)
{
	delete ((char *)(element.get()));
}

void deleteComments(PointerArray * comments)
{
	if (!comments) return;
	comments->doAll(killComment, NULL);
	delete comments;
}

InitFile::InitFile(char * filenm) : changed(FALSE)
{
	ifstream ini(filenm);
	unsigned char first;
	PointerArray * sectionComments = NULL;

	setName(filenm);
	if (!ini) {
		cerr << "Cannot open ini file." << endl;
		return;
	}
	// Read in the file contents.
	data.put((Section::pointers).getLastIndex());
	while (!ini.eof()) {
		ini >> first;
		if ((first == '[') || (first == ';')) {
			// Section heading.
			ini.putback(first);
			Section * temp = new Section;
			sectionComments = ((InitFile *)temp)->insert(ini,
				sectionComments, data.getInt());
			++ len;
		}
		else {
			// I'm not interested.
			streamIgnore(ini);
		}
	}
	if (!(len.getInt())) comments = sectionComments;
	pointers.searchAndInsert(0, Key::compare, this);
}

PointerArray Section::pointers;

PointerArray * Section::insert(istream & ini,
	PointerArray * sectionComments, int where)
{
	unsigned char first;
	PointerArray * retval = NULL;

	// Get section comments - only if we don't already have them.
	if (!sectionComments)
		comments = ::getComments(ini);
	else comments = sectionComments;
	ini >> first;
	while ((first != '[') && !(ini.eof())) {
		// This happens in the case of assigments at the top
		// of a .INI file, before a section heading.
		// The assignments, and any comments after the
		// assignments, get ignored (until the first
		// section heading.)
		// If you're really picky, I guess I'd agree that
		// there's a better way to do this.
		streamIgnore(ini);
		ini >> first;
	}
	if (ini.eof()) return retval;
	// Read in the name of the section.
	unsigned char buffer[INI_LINESIZE + 1];
	char * spot = getline(ini, buffer);
	if (*spot != ']') {
		cerr << "Error in ini file." << endl;
		return retval;
	}
	*spot = '\0';
	setName(buffer);
	data.put((Key::pointers).getLastIndex());
	// Read in the section contents.
	ini >> first;
	ini.putback(first);
	while (!ini.eof()) {
		if (first == '[') {
			// Next section heading - we're done with this one.
			break;
		}
		else {
			// Assume it's an 'assignment.'
			Key * assign = new Key(ini, data.getInt());
			if (!(assign->nameOf())) {
				// next section.
				retval = assign->getComments();
				// The return value should be passed
				// into the next section's insert().
				assign->clearComments();
				delete assign;
				break;
			}
			++ len;
		}
		ini >> first;
		ini.putback(first);
	}
	pointers.searchAndInsert(where, compare, this);
	return retval;
}

PointerArray Key::pointers(100);

Key::Key(istream & ini, int where) : name(NULL), comments(NULL)
{
	unsigned char buffer[INI_LINESIZE + 1];
	// Read in the 'assignment.'
	if (!ini.eof()) {
		comments = ::getComments(ini);
		if (ini.eof()) return;
		ini >> buffer[0];
		if (buffer[0] == '[') {
			ini.putback(buffer[0]);
			return;
		}
		// For now, only INI_LINESIZE size assignments allowed.
		getline(ini, buffer + 1);
		char * equals = strchr(buffer, '=');
		*equals = '\0';
		setName(buffer);
		(getData()).putString(equals + 1);
		pointers.searchAndInsert(where, compare, this);
	}
}

// Following makes a COPY of the argument.
char * Key::setName(char * value)
{
	name = new char[strlen(value) + 1];
	strcpy(name, value);
	return name;
}

int PointerArray::newElement()
{
	if (!(numFilled % chunkSize) && numFilled) {
		AbsPtr * biggerArea = new AbsPtr[numFilled + chunkSize];
		for (int i = 0; i < numFilled; i ++)
			biggerArea[i] = ptrArray[i];
		delete [] ptrArray;
		ptrArray = biggerArea;
	}
	int retval = getLastIndex();
	numFilled ++;
	return retval;
}

int Key::compare(void * val1, void * val2)
{
	return (strcmp(((Key *)val1)->name, ((Key *)val2)->name));
}

AbsPtr &PointerArray::bumpUp(int position, int last)
{
	AbsPtr * current = &ptrArray[position];
	// bump up elements
	for (AbsPtr * temp = &ptrArray[last - 1]; temp >= current; temp --)
		(temp[1]).put(temp->get());
	return *current;
}

void PointerArray::searchAndInsert(int where,
    int (*compare)(void *, void *), void * element)
{
	int last = newElement(), position;
	// Perform insertion.
	if ((position = search(element, where, last, compare)) < 0)
		position = -(position + 1);
	bumpUp(position, last).put(element);
}

PointerArray::insert(int where, void * element)
{
	int last = newElement();
	bumpUp(where, last).put(element);
	return last;
}

void PointerArray::insertAtEnd(void * value)
{
	int last = newElement();
	(ptrArray[last]).put(value);
}

void PointerArray::insertString(char * value)
{
	int last = newElement();
	(ptrArray[last]).putString(value);
}

// The following search has two cases for return:
// If an exact match of element is found, the index of the
// matched element is returned.  Otherwise, the index of the
// closest element lower than the argument is returned, with
// negative sign to indicate that there was no exact match.
// (Ooops, since there is no negative zero, one will be
// subtracted in the case of no match.)
// I am assuming the upperBound is EXCLUSIVE.  There are
// no elements to search if lowerBound == upperBound, so
// -lowerBound - 1 will be returned.
int PointerArray::search(void * element, int lowerBound, int upperBound,
	int (*compare)(void *, void *))
{
	// Perform a binary search.
	if (upperBound == lowerBound) return (-(++lowerBound));
	upperBound --;
	int current, found = FALSE, result;
	do {
		current = (lowerBound + upperBound) / 2;
		if ((result = compare(element,
			(ptrArray[current]).get())) < 0) {
			// lower the bounds
			upperBound = current - 1;
		}
		else {
			// else raise the bounds
			lowerBound = current + 1;
		}
		if (!result) found = TRUE;
	} while ((!found) && (upperBound >= lowerBound));
	if (found) return current;
	return (-lowerBound - 1);
}

ostream & Key::dump(ostream & where)
{
	dumpComments(where, comments);
	where << nameOf() << "=" << valueOf() << endl;
	return where;
}

ostream & InitFile::dump(ostream & where)
{
	dumpComments(where, comments);
	int start = data.getInt();
	int end = start + len.getInt();
	for (int i = start; i < end; i ++) {
		((Section *)((Section::pointers)[i]))->dump(where);
	}
	return where;
}

ostream & Section::dump(ostream & where)
{
	if (comments) dumpComments(where, comments);
	where << "[" << nameOf() << "]" << endl;
	int start = getData().getInt();
	int end = start + len.getInt();
	for (int i = start; i < end; i ++) {
		((Key *)((Key::pointers)[i]))->dump(where);
	}
	where << endl;
	return where;
}

void PointerArray::doAll(void (*func)(AbsPtr & element, void * userPtr),
	void * userPtr)
{
	for (int i = 0; i < numFilled; i ++) func(ptrArray[i], userPtr);
}

int InitFile::compare(void * val1, void * val2)
{
	return (strcmp((char *)val1, ((Section *)val2)->nameOf()));
}

int Section::compare2(void * val1, void * val2)
{
	char * str2 = ((Key *)val2)->nameOf();
	// If str2 is NULL, a NEGATIVE value should be returned.
	// NULL should be considered the heaviest string.
	// This occurs when we are adding a new key.
	return (str2 ? strcmp((char *)val1,str2) : -1);
}

Key * Section::findAndCreate(PointerArray & ptrs,
	int (*compareFunc)(void *, void *), PointerArray & topPtrs,
	Key * (*makeFunc)(char *, char *), char * lval, char * rval,
	int & changed)
{
	int start = data.getInt();
	int end = start + len.getInt();
	int found = ptrs.search(lval, start, end, compareFunc);
	if (found < 0) {
		if (!rval) return NULL;
		// Create and insert a new key.
		changed = TRUE;
		ptrs.insert(found = -(found + 1), makeFunc(lval, rval));
		// Problem here.  Suppose other sections come
		// after this one in Section::pointers.
		// They need to have their 'data' incremented.
		int last = topPtrs.getLastIndex();
		for (int i = 0; i < last; i ++) {
			AbsPtr & cur = ((Section *)(topPtrs[i]))->data;
			int pos = cur.getInt();
			if ((pos != start) && (pos >= found)) ++ cur;
		}
		++ len;
	}
	Key * foundKey = (Key *)(ptrs[found]);
	return foundKey;
}

char * Section::find(char * lval, char * rval, int & changed)
{
	Key * found = findAndCreate(Key::pointers, compare2,
		pointers, Key::maker, lval, rval, changed);
	return (found ? found->valueOf() : NULL);
}

char * InitFile::find(char * section, char * lval, char * rval)
{
	Key * found = findAndCreate(Section::pointers, InitFile::compare,
		pointers, Section::maker, section, rval, changed);
	return (found ? ((InitFile *)found)->
		Section::find(lval, rval, changed) : NULL);
}

int InitFile::find(char * section, char * lval, int rval)
{
	char arval[20];
	return (atoi(find(section, lval, itoa(rval, arval, 10))));
}

#pragma argsused
Section::Section(char * sectionName, char * dummy) : len(0)
{
	data.put((Key::pointers).getLastIndex());
	setName(sectionName);
}

char * AbsPtr::putString(char * value)
{
	char * copy = new char[strlen(value) + 1];
	strcpy(copy, value);
	put(copy);
	return copy;
}

PointerArray InitFile::pointers(3);

void Section::destructor(PointerArray & ptrs, void (*killFunc)(void *))
{
//	deleteComments(comments);
//	if (name) delete name;
	int start = data.getInt();
	int end = start + len.getInt();
	for (int i = start; i < end; i ++)
		killFunc(ptrs[i]);
}

Key::~Key()
{
	deleteComments(comments);
	if (name) delete name;
}

void InitFile::write(char * filename)
{
	ofstream * where = NULL;
	if ((!filename) && changed) where = new ofstream(name);
	if (filename) where = new ofstream(filename);
	if (where) {
		dump(*where);
		delete(where);
	}
}
